#!/usr/bin/env python

# split_on_guides.py - a Python-fu script which splits an image along its guides
# Courtesy of Ice Pillar Computing
# http://www.icepillar.com/freesw/split-on-guides/
# Version 1.2
#
# Copyright (C) 2003  Dwayne C. Litzenberger <dlitz@dlitz.net>
#
# Early Public Domain Release Notice:
# * This work shall enter the public domain on the first day of the tenth
# * (10th) year following the latest copyright year listed above.  Until then,
# * the following terms shall apply:
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# or visit http://www.fsf.org/licenses/gpl.html
#
# History:
#   1.0     30 August 2003
#       - Initial Release
#
#   1.1     30 August 2003
#       - Added 'border="0"' to <table> tag
#
#   1.2     31 August 2003
#       - The "I Really Mean It This Time" release
#       - Major code cleanup
#       - No longer touches the source image or the clipboard.  This comes at
#         the expense of memory usage, since we now duplicate the entire image
#         before doing anything.
#       - Added support for multi-layer images.
#       - Worked around "quantization tables are too coarse for baseline JPEG"
#         problem with JPEG images
#       - Removed support for comma separator in file_extenstions
#       - Added option for JPEG quality setting

# file_jpeg_save parameters
# Used to work around the problem related to the "Caution: quantization tables
# are too coarse for baseline JPEG" warning message.
NORMAL_JPEG_PARAMS = (
    0.75,       # quality
    0.0,        # smoothing
    True,       # optimize
    False,      # progressive
    "Created with The GIMP",   # comment
    0,          # subsmp (0 "2x2, 1x1, 1x1")
    False,      # baseline
    0,          # restart
    1)          # dct (1 means "Integer")

HIGH_QUALITY_JPEG_PARAMS = (
    1.0,       # quality
    0.0,        # smoothing
    True,       # optimize
    False,      # progressive
    "Created with The GIMP",   # comment
    2,          # subsmp (2 means "1x1, 1x1, 1x1")
    False,      # baseline
    0,          # restart
    2)          # dct (2 means "Floating-Point")

#### Init ####

from gimpfu import *
import os
import sys
import cgi, urlparse

#### Some Python < 2.3 compatibility stuff ####
try:
    from sets import Set
except ImportError:
    class Set:
        """Partial implementation of Python 2.3's Set type"""
        def __init__(self):
            self.data = {}
        def add(self, item):
            self.data[item] = None
        def remove(self, item):
            if self.data.has_key(item):
                del self.data[item]
        def __repr__(self):
            return "Set(" + repr(self.data.keys()) + ")"
        def __getitem__(self, index):
            return self.data.keys()[index]
        def __len__(self):
            return len(self.data.keys())

if not hasattr(__builtins__, 'enumerate'):
    def enumerate(iterable):
        """Partial implementation of Python 2.3's enumerate() builtin"""
        i = 0
        retval = []
        for v in iterable:
            retval.append((i, v))
            i += 1
        return retval


#### Helper functions ####

class GimpExceptionWrapper:
    """GimpExceptionWrapper(callable) -> callable

When called, any exception which is raised when calling callable results in a
GIMP message being displayed to the user.
"""
    def __init__(self, callable):
        self.callable = callable
    def __call__(self, *args, **kwargs):
        try:
            return self.callable(*args, **kwargs)
        except:
            pdb.gimp_message(
                str(sys.exc_info()[0]) + ": " + str(sys.exc_info()[1]))
            raise


def guides_to_rectangles(image):
    """guides_to_rectangles(image) -> [row1, row2, ...]

Return a list of rows of rectangles defined by the guides in image.

Each row is a list of tuples: (offset_x, offset_y, width, height)
"""
    # Load the guides into their corresponding sets
    h_guides = Set()
    v_guides = Set()
    guide = image.find_next_guide(0)
    if not guide:
        raise ValueError, "image must contain at least one guide"
    while guide:
        orientation = image.get_guide_orientation(guide)
        position = image.get_guide_position(guide)
        (h_guides, v_guides)[orientation].add(position)
        guide = image.find_next_guide(guide)
    h_guides.add(image.height)
    v_guides.add(image.width)
    
    # Remove any zero entry, since zero is implied
    h_guides.remove(0)
    v_guides.remove(0)

    # Convert the sets into lists so we can sort them
    h_guides = list(h_guides)
    v_guides = list(v_guides)
    
    # Sort the lists
    h_guides.sort()
    v_guides.sort()
    
    # Build the rectangles
    rows = []
    y1 = 0
    for y2 in h_guides:
        x1 = 0
        cols = []
        for x2 in v_guides:
            cols.append((x1, y1, x2 - x1, y2 - y1))
            x1 = x2
        rows.append(cols)
        y1 = y2
    
    return rows

#### The Python-Fu function itself ####

def split_on_guides(timg, tdrawable, leave_open=True, save_files=False,
                    image_directory=None, filename_template=None,
                    file_extensions=None, jpeg_quality=0.75, 
                    save_html=False, html_filename=None,
                    image_baseurl="", save_info=False, info_filename=None,
                    merge_layers=True):
    # Fix the image_baseurl parameter so urljoin does what we expect
    if image_baseurl and not image_baseurl.endswith("/"):
        image_baseurl += "/"
    
    # Parse the jpeg_quality parameter
    if jpeg_quality > 1.0:
        jpeg_params = HIGH_QUALITY_JPEG_PARAMS
    else:
        jpeg_params = list(NORMAL_JPEG_PARAMS)
        jpeg_params[0] = jpeg_quality

    # Split file_extensions string into a list
    if file_extensions:
        if " " in file_extensions:
            file_extensions = file_extensions.split(" ")
        else:
            file_extensions = [file_extensions]
        # Remove blank entries
        file_extensions = filter(None, [x.strip() for x in file_extensions])
    else:
        file_extensions = []
    
    # Sanity checks
    if not (leave_open or save_files or save_html or save_info):
        raise TypeError, "cowardly refusing to do nothing"
    if (save_files or save_html or save_info):
        if not file_extensions:
            raise TypeError, \
                "save_files, save_html, and save_info require" + \
                "at least one file extension"
        if not filename_template:
            raise TypeError, \
                "info_filename, save_html and save_files require" + \
                "a filename template"
    
    # Get the rectangles
    rows = guides_to_rectangles(timg)
    numrows = len(rows)
    numcols = len(rows[0])
    total_images = numrows * numcols

    # Open output files
    if save_html:
        html_file = open(html_filename, "w")
        html_file.write('<table cellpadding="0" cellspacing="0" border="0">')
    if save_info:
        info_file = open(info_filename, "w")
        
    # Initialize progress bar
    try: gimp.progress_init("Splitting...")
    except: pass    # TODO - Is this really needed for non-interactive mode?
    progress = 0

    # Main loop
    for rownum, row in enumerate(rows):
        if save_html: html_file.write('<!--\nNew Row --><tr>')
        for colnum, (x, y, w, h) in enumerate(row):
            # Create the new image
            newimg = pdb.gimp_channel_ops_duplicate(timg)
            pdb.gimp_crop(newimg, w, h, x, y)
            
            # Get the new drawable
            layer_count = pdb.gimp_image_get_layers(newimg)[0]
            if merge_layers and layer_count > 1:
                newdrawable = pdb.gimp_image_merge_visible_layers(newimg,
                    EXPAND_AS_NECESSARY)
            else:
                newdrawable = pdb.gimp_image_get_active_layer(newimg)
            
            for i, ext in enumerate(file_extensions):
                filename = filename_template % {
                            'row': rownum,
                            'col': colnum,
                            'ext': ext,
                        }

                # Save the image
                if save_files:
                    if ext == "jpg" or ext == "jpeg":
                        pdb.file_jpeg_save(
                            newimg,
                            newdrawable,
                            os.path.join(image_directory, filename),
                            filename,
                            *jpeg_params)
                    else:
                        pdb.gimp_file_save(
                            newimg,
                            newdrawable,
                            os.path.join(image_directory, filename),
                            filename)
                    if i == 0:
                        pdb.gimp_image_clean_all(newimg)
                        pdb.gimp_image_set_filename(
                            newimg, os.path.join(image_directory, filename))
                
                # Write the image info
                if save_info:
                    info_file.write(
                        "%d %d %d %d %s\n" % (rownum, colnum, w, h, filename))

                # Write the HTML data, using the first extension
                if save_html and i == 0:
                    # The HTML code is intentionally this ugly so that there is
                    # never any whitespace in between the elements (some browsers
                    # may insert whitespace between the images if there is
                    # whitespace between the elements).
                    html_file.write(
                        '''<td
                            width="%(width)s" height="%(height)s"
                        ><!--<a href="">--><img
                            src="%(src)s"
                            alt=""
                            width="%(width)s" height="%(height)s"
                            border="0"
                        /><!--</a>--></td>''' % {
                            'width': str(w),
                            'height': str(h),
                            'src':
                                cgi.escape(
                                    urlparse.urljoin(image_baseurl, filename)),
                        })
            
            
            # Depending on the value of leave_open, either open a new window
            # for the new image, or delete the image from memory.
            if leave_open:
                pdb.gimp_display_new(newimg)
            else:
                pdb.gimp_image_delete(newimg)
            
            # Update progress bar
            progress += 1
            try: gimp.progress_update(float(progress) / total_images)
            except: pass    # TODO - See comment above for progress_init
        if save_html: html_file.write('''</tr>''')
    
    # Cleanup files
    if save_html:
        html_file.write("</table>")
        html_file.close()
    if save_info:
        info_file.close()

    # Done.


#### Plugin initialization ####

register(
    "python_fu_split_on_guides",            # name
    "Split an image along its guides",      # blurb
    "Split an image along its guides",      # help
    "Dwayne C. Litzenberger",               # author
    "Dwayne C. Litzenberger",               # copyright
    "2003",                                 # date
    "<Image>/Guides/Split on Guides",       # menupath
    "*",                                    # imagetypes
    [                                       # params
        (PF_BOOL, "leave_open", "Leave images open?", False),
        (PF_BOOL, "save_files", "Save files?", False),
        (PF_STRING, "image_directory", "Image directory",
            os.path.join(os.getcwd(), "images")),
        (PF_STRING, "filename_template",
            "Image filename template\n" + 
                "%(row)s - row number\n" +
                "%(col)s - column number\n" +
                "%(ext)s - file extension",
            "img_%(row)s_%(col)s.%(ext)s"),
        (PF_STRING, "file_extensions",
            "File types to save, separated by spaces\n" +
                "The first extension will be used for the HTML table.",
            'png'),
        (PF_FLOAT, "jpeg_quality",
            "JPEG quality\n" +
            "Setting this to a value > 1.0 causes" +
            "the subsampling and DCT method to be changed as well\n",
            0.75),
        (PF_BOOL, "save_html", "Save an HTML table?", False),
        (PF_STRING, "html_filename", "Filename for HTML file",
            os.path.join(os.getcwd(), "images.html")),
        (PF_STRING, "image_baseurl", "Base image URL for HTML file", "images"),
        (PF_BOOL, "save_info",
            "Save an image info file?\n" +
                "Consists of one image per line:\n" +
                "row col width height filename",
            False),
        (PF_STRING, "info_filename", "Filename for info file",
            os.path.join(os.getcwd(), "images.txt")),
        (PF_BOOL, "merge_layers",
            "Merge visible layers for the resulting image?\n" +
            "Usually, you want this unless you're saving XCF format files.",
            True),
    ],
    [],                                     # results
    GimpExceptionWrapper(split_on_guides))  # function

main()

# vim:set ts=4 sw=4 sts=4 expandtab:    # Using emacs is silly!
